#include "atlas_misc.h"
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_blk.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_ablk2cmat.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_a1.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_an.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_aX.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_flag.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_kern.h))
#include "atlas_lvl2.h"
#include "atlas_level1.h"
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_rankK.h))

/*
 * This routine attempt to choose which access-major GEMM to call based on dims
 */
int Mjoin(PATL,ammm)
(
   enum ATLAS_TRANS TA,
   enum ATLAS_TRANS TB,
   ATL_CSZT M,
   ATL_CSZT N,
   ATL_CSZT K,
   const SCALAR alpha,
   const TYPE *A,
   ATL_CSZT lda,
   const TYPE *B,
   ATL_CSZT ldb,
   const SCALAR beta,
   TYPE *C,
   ATL_CSZT ldc
)
{
   int iret;
/*
 * Just do a scale and return
 */
   if (alpha == ATL_rzero)
   {
      if (beta == ATL_rzero)
         Mjoin(PATL,gezero)(M, N, C, ldc);
      else if (beta != ATL_rone)
         Mjoin(PATL,gescal)(M, N, beta, C, ldc);
      return(0);
   }
   if (K == 1)  /* really a GER */
   {
      if (!SCALAR_IS_ONE(beta))
      {
         int i;
         ATL_CSZT incA = (TA == AtlasNoTrans) ? 1 : lda;
         ATL_CSZT incB = (TB == AtlasNoTrans) ? ldb : 1;
         for (i=0; i < N; i++, B += incB, C += ldc)
            Mjoin(PATL,axpby)(M, alpha * *B,  A, incA, beta, C, 1);
      }
      else
         Mjoin(PATL,ger)(M, N, alpha, A, (TA == AtlasNoTrans) ? 1 : lda,
                         B, (TB == AtlasNoTrans) ? ldb : 1, C, ldc);
      return(0);
   }
   else if (K == 2 && SCALAR_IS_ONE(beta))  /* really a GER2 */
   {
      if (TA == AtlasNoTrans)
      {
         if (TB == AtlasNoTrans)
            Mjoin(PATL,ger2)(M, N, alpha, A, 1, B, ldb, alpha, A+lda, 1,
                             B+1, ldb, C, ldc);
         else
            Mjoin(PATL,ger2)(M, N, alpha, A, 1, B, 1, alpha, A+lda, 1,
                             B+ldb, 1, C, ldc);
      }
      else if (TB == AtlasNoTrans)
         Mjoin(PATL,ger2)(M, N, alpha, A, lda, B, ldb, alpha, A+1, lda,
                          B+1, ldb, C, ldc);
      else
         Mjoin(PATL,ger2)(M, N, alpha, A, lda, B, 1, alpha, A+1, lda,
                          B+ldb, 1, C, ldc);
      return(0);
   }
   else if (N == 1)  /* Really GEMV with A as matrix, A & C as vectors */
   {
      if (TA == AtlasNoTrans)
         Mjoin(PATL,gemv)(AtlasNoTrans, M, K, alpha, A, lda, B,
                          (TB == AtlasNoTrans) ? 1:ldb, beta, C, 1);
      else
         Mjoin(PATL,gemv)(AtlasTrans, K, M, alpha, A, lda, B,
                          (TB == AtlasNoTrans) ? 1:ldb, beta, C, 1);
      return(0);
   }
   else if (M == 1)  /* Really GEMV with B as matrix, A & C as vectors */
   {
      if (TB == AtlasNoTrans)
         Mjoin(PATL,gemv)(AtlasTrans, K, N, alpha, B, ldb, A,
                          (TA == AtlasNoTrans) ? lda:1, beta, C, ldc);
      else
         Mjoin(PATL,gemv)(AtlasNoTrans, N, K, alpha, B, ldb, A,
                          (TA == AtlasNoTrans) ? lda:1, beta, C, ldc);
      return(0);
   }
   else if (M <= ATL_AMM_MAXMB && N <= ATL_AMM_MAXNB && K <= ATL_AMM_MAXKB)
      iret = Mjoin(PATL,ammm_1b)(TA, TB, M, N, K, alpha, A, lda, B, ldb,
                                 beta, C, ldc);
   else if (K <= ATL_MAXK_RKK)
      iret = Mjoin(PATL,ammm_rkK)(TA, TB, M, N, K, alpha, A, lda, B, ldb,
                                  beta, C, ldc);
   else if (M >= 3*ATL_AMM_MAXMB && N >= 3*ATL_AMM_MAXNB &&
            K >= 3*ATL_AMM_MAXKB)
      iret = Mjoin(PATL,ammmNMK)(ATL_AMM_NCASES-1, TA, TB, M, N, K, alpha,
                                 A, lda, B, ldb, beta, C, ldc);
   else if (K < ATL_AMM_MAXKB)
   {
      int i;
      for (i=0; i < ATL_AMM_NCASES && ATL_AMM_KBs[i] < K; i++);
      iret = Mjoin(PATL,ammmNMK)(i, TA, TB, M, N, K, alpha, A, lda,
                                 B, ldb, beta, C, ldc);
   }
   else
      iret = Mjoin(PATL,ammmNMK)(ATL_AMM_NCASES-1, TA, TB, M, N, K, alpha,
                                 A, lda, B, ldb, beta, C, ldc);
   return(iret);
}
