#include "atlas_misc.h"
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_blk.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_ablk2cmat.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_a1.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_flag.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_kern.h))

int Mjoin(PATL,ammmNMK)
(
   int IK,                /* index of kernel to use */
   enum ATLAS_TRANS TA,
   enum ATLAS_TRANS TB,
   ATL_CSZT M,
   ATL_CSZT N,
   ATL_CSZT K,
   const SCALAR alpha,
   const TYPE *A,
   ATL_CSZT lda,
   const TYPE *B,
   ATL_CSZT ldb,
   const SCALAR beta,
   TYPE *C,
   ATL_CSZT ldc
)
{
   ATL_CINT mu=ATL_AMM_MUs[IK], nu=ATL_AMM_NUs[IK], ku=ATL_AMM_KUs[IK];
   #if ATL_AMM_MAXKMAJ > 1
      size_t kb0U, KK;
   #else
      #define kb0U kb0
      #define KK K
   #endif
   ATL_CINT KRUN = ATL_AMM_KRUNTIME(IK);
   ATL_CSZT mb=ATL_AMM_MBs[IK], NB=ATL_AMM_NBs[IK], kb=ATL_AMM_KBs[IK];
   ATL_CSZT incwA = mb*kb, nmu = mb / mu;
   size_t nmblks, nnblks, nkblks, mbF, nbF, kb0, nmuF, MBF, NBF;
   size_t incAk, incBk, incAk0, incBk0, incAm, incBn, incCn;
   size_t i, j, k, szA, szB, szC, nnu0, nnuF;
   const TYPE *B0 = B;
   TYPE *wA, *wB, *wC, *wA0, *wB0;
   void *vp;
   cm2am_t a2blk, b2blk;
   ablk2cmat_t blk2c;
   ammkern_t amm_b1 = ATL_AMM_KERN_b1[IK], amm_b0;

   if (alpha == ATL_rone)
   {
      if (beta == ATL_rone)
         blk2c = ATL_AMM_BLK2C_a1_b1[IK];
      else if (beta == ATL_rnone)
         blk2c = ATL_AMM_BLK2C_a1_bn[IK];
      else if (beta == ATL_rzero)
         blk2c = ATL_AMM_BLK2C_a1_b0[IK];
      else
         blk2c = ATL_AMM_BLK2C_a1_bX[IK];
   }
   else if (alpha == ATL_rnone)
   {
      if (beta == ATL_rone)
         blk2c = ATL_AMM_BLK2C_an_b1[IK];
      else if (beta == ATL_rnone)
         blk2c = ATL_AMM_BLK2C_an_bn[IK];
      else if (beta == ATL_rzero)
         blk2c = ATL_AMM_BLK2C_an_b0[IK];
      else
         blk2c = ATL_AMM_BLK2C_an_bX[IK];
   }
   else
   {
      if (beta == ATL_rone)
         blk2c = ATL_AMM_BLK2C_aX_b1[IK];
      else if (beta == ATL_rnone)
         blk2c = ATL_AMM_BLK2C_aX_bn[IK];
      else if (beta == ATL_rzero)
         blk2c = ATL_AMM_BLK2C_aX_b0[IK];
      else
         blk2c = ATL_AMM_BLK2C_aX_bX[IK];
   }
/*
 * Handle N differently than other dims: since it is outer loop, don't want
 * to peel, so just vary nb inside the main loop.  nnblks therefore includes
 * final block, unlike for M or K.
 */
   if (N >= NB+nu+nu)
   {
      nnblks = N/NB;
      nbF = N - nnblks * NB;
      if (nbF < nu+nu)
         nbF += NB;
      else
         nnblks++;
   }
   else
   {
      nnblks = 1;
      nbF = N;
   }
   nnu0 = NB / nu;
   nnuF = (nbF+nu-1)/nu;
   NBF = nnuF * nu;
/*
 * For M, we must peel the final block to handle any cleanup (can't peel 1st
 * block or we mess up alignment!), so this block is not included in the
 * block count
 */
   if (M >= mb+mu+mu)  /* more than just last block */
   {
      nmblks = M/mb;
      mbF = M - nmblks * mb;
      if (mbF < mu+mu)  /* steal block from main iteration, not enough here! */
      {
         nmblks--;
         mbF += mb;
      }
   }
   else /* put everything in final block */
   {
      mbF = M;
      nmblks = 0;
   }
   nmuF = (mbF+mu-1) / mu;
   MBF = nmuF * mu;
/*
 * For K, we peel the first iteration to set BETA=0, so the nkblks does not
 * include the peeled block
 */
   if (K >= kb)
   {
      nkblks = K/kb;
      kb0 = K - nkblks * kb;
      if (!kb0)
      {
         kb0 = kb;
         nkblks--;
      }
      else
      {
         if (kb0 < 4)
         {
            kb0 += kb;
            nkblks--;
         }
      }
   }
   else /* K < nb */
   {
      kb0 = K;
      nkblks = 0;
   }
   #if ATL_AMM_MAXKMAJ > 1
      if (ATL_AMM_KMAJOR(IK))
      {
         KK = ((K+ku-1)/ku)*ku;
         kb0U = ((kb0+ku-1)/ku)*ku;
         if (ATL_AMM_KRUNTIME(IK))
            amm_b0 = ATL_AMM_KERN_b0[IK];
         else
            amm_b0 = (ATL_AMM_KBs[IK] == kb0U)
                     ?  ATL_AMM_KERN_b0[IK] : ATL_AMM_KERN_K1[IK];
      }
      else
      {
         KK = K;
         kb0U = kb0;
         amm_b0 = (kb0==ATL_AMM_KBs[IK] || ATL_AMM_KRUNTIME(IK))
                  ? ATL_AMM_KERN_b0[IK]:ATL_AMM_KERN_K1[IK];
      }
   #else
      amm_b0 = (kb0==ATL_AMM_KBs[IK] || ATL_AMM_KRUNTIME(IK))
               ? ATL_AMM_KERN_b0[IK]:ATL_AMM_KERN_K1[IK];
   #endif
   szA = (nmblks*mb+MBF)*KK; /* wrkspc for all of A wt M rounded up to MU*/
   j = Mmax(NB, NBF);
   i = Mmax(MBF, mb);
   szC = i*j;
   #if ATL_AMM_MAXDUPB < 2
      szB = KK*j;                     /* workspace for panel of B */
   #else
      szB = (KK*j)*ATL_AMM_DUPBN[IK]; /* workspace for panel of B */
   #endif

   vp = malloc(ATL_MulBySize(szA+szB+szC)+3*ATL_Cachelen);
   ATL_assert(vp);
   wB0 = wB = ATL_AlignPtr(vp);
   wA = wB + szB;
   wA0 = wA = ATL_AlignPtr(wA);
   wC = wA + szA;
   wC = ATL_AlignPtr(wC);

   if (TA == AtlasNoTrans)
   {
      incAm = mb;
      incAk0 = kb0*lda;
      incAk = kb*lda;
      a2blk = ATL_AMM_AT2BLK_a1[IK];
   }
   else
   {
      incAm = mb*lda;
      incAk0 = kb0;
      incAk = kb;
      a2blk = ATL_AMM_A2BLK_a1[IK];
   }
   if (TB == AtlasNoTrans)
   {
      incBk0 = kb0;
      incBk = kb;
      b2blk = ATL_AMM_B2BLK_a1[IK];
      incBn = NB*ldb;
   }
   else
   {
      incBk0 = kb0*ldb;
      incBk = kb*ldb;
      b2blk = ATL_AMM_BT2BLK_a1[IK];
      incBn = NB;
   }
   incCn = ldc*NB;

   for (j=0; j < nnblks; j++)
   {
      size_t nb, nbsz, incwB, incwB0, nnu;
      const TYPE *Bn = B+incBn;
      TYPE *Cn = C + incCn;
      if (j != nnblks-1)
      {
         nbsz = nb = NB;
         nnu = nnu0;
      }
      else
      {
         nb = nbF;
         nbsz = NBF;
         nnu = nnuF;
      }
      incwB = nbsz*kb;
      incwB0 = nbsz*kb0U;
      #if ATL_AMM_MAXDUPB > 1
         incwB  *= ATL_AMM_DUPBN[IK];
         incwB0 *= ATL_AMM_DUPBN[IK];
      #endif
/*
 *    Do all M-blocks except final one, which may be of differing size & partial
 */
      for (i=0; i < nmblks; i++)
      {
         TYPE *wAn, *wBn;
         const TYPE *An = A+incAm;
/*
 *       Peel first K it to handle K-cleanup and set BETA=0
 */
         if (!j)
            a2blk(kb0, mb, ATL_rone, A, lda, wA);
         if (!i)
            b2blk(kb0, nb, ATL_rone, B, ldb, wB);
         wAn = wA+mb*kb0U;
         wBn = (nkblks) ? wB+incwB0 : wB;
         amm_b0(nmu, nnu, kb0U, wA, wB, wC, nkblks?wAn:wA, wBn, wC);
         wA = wAn;
         wB = wBn;
         A += incAk0;
         B += incBk0;
/*
 *       If first K-block not the only K-block
 */
         if (nkblks)
         {
            for (k=nkblks-1; k; k--)
            {
               wAn = wA+incwA;
               wBn = wB+incwB;
               if (!j)
                  a2blk(kb, mb, ATL_rone, A, lda, wA);
               if (!i)
                  b2blk(kb, nb, ATL_rone, B, ldb, wB);
               amm_b1(nmu, nnu, kb, wA, wB, wC, wAn, wBn, wC);
               wA = wAn;
               wB = wBn;
               A += incAk;
               B += incBk;
            }
/*
 *          Last K-block peeled to change prefetch pattern
 */
            wAn = wA+incwA;
            if (!j)
               a2blk(kb, mb, ATL_rone, A, lda, wA);
            if (!i)
               b2blk(kb, nb, ATL_rone, B, ldb, wB);
            amm_b1(nmu, nnu, kb, wA, wB, wC, wAn, wB0, wC);
            wA = wAn;
            wB = wB0;
         }
         blk2c(mb, nb, alpha, wC, beta, C, ldc);
         A = An;
         B = B0;
         C += mb;
      }
/*
 *    Do the final peeled M-block, which is of non-constant size mbF
 */
      {
         TYPE *wAn, *wBn;
/*
 *       Peel first K it to handle K-cleanup and set BETA=0
 */
         if (!j)
            a2blk(kb0, mbF, ATL_rone, A, lda, wA);
         if (!i)
            b2blk(kb0, nb, ATL_rone, B, ldb, wB);
         wAn = wA+MBF*kb0U;
         wBn = (nkblks) ? wB+incwB0 : wB;
         amm_b0(nmuF, nnu, kb0U, wA, wB, wC, nkblks?wAn:wA, wBn, wC);
         wA = wAn;
         wB = wBn;
         A += incAk0;
         B += incBk0;
/*
 *       If first K-block not the only K-block
 */
         if (nkblks)
         {
            for (k=nkblks-1; k; k--)
            {
               wAn = wA+MBF*kb;
               wBn = wB+incwB;
               if (!j)
                  a2blk(kb, mbF, ATL_rone, A, lda, wA);
               if (!i)
                  b2blk(kb, nb, ATL_rone, B, ldb, wB);
               amm_b1(nmuF, nnu, kb, wA, wB, wC, wAn, wBn, wC);
               wA += MBF*kb;
               wB = wBn;
               A += incAk;
               B += incBk;
            }
/*
 *          Last K-block peeled to change prefetch pattern
 */
            if (!j)
               a2blk(kb, mbF, ATL_rone, A, lda, wA);
            if (!i)
               b2blk(kb, nb, ATL_rone, B, ldb, wB);
            amm_b1(nmuF, nnu, kb, wA, wB, wC, wA0, wB0, wC);
            wA = wA0;
            wB = wB0;
         }
         blk2c(mbF, nb, alpha, wC, beta, C, ldc);
      }  /* end M-peel */
      wA = wA0;
      B = B0 = Bn;
      C = Cn;
   }
   free(vp);
   return(0);
}
