#include "atlas_misc.h"
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_blk.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_ablk2cmat.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_a1.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_an.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_cm2am_aX.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_flag.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_kern.h))
/*
 * This routine called in degenerate case where all dims less than max block,
 * so we can do entire operation with one kernel call
 */
int Mjoin(PATL,ammm_1b)
(
   enum ATLAS_TRANS TA,
   enum ATLAS_TRANS TB,
   ATL_CSZT M,
   ATL_CSZT N,
   ATL_CSZT K,
   const SCALAR alpha,
   const TYPE *A,
   ATL_CSZT lda,
   const TYPE *B,
   ATL_CSZT ldb,
   const SCALAR beta,
   TYPE *C,
   ATL_CSZT ldc
)
{
   int i;
   int nmu, nnu, nku, bM, bN, bK, szB;
   #if ATL_AMM_MAXKMAJ > 1
      int KK;
   #else
      #define KK K
   #endif
   int mu, nu, ku;
   void *vp;
   TYPE *pA, *pB, *pC, *p;
   TYPE alpA=ATL_rone, alpB=ATL_rone, alpC=ATL_rone;
   ammkern_t amm;
   ablk2cmat_t blk2c;
   cm2am_t a2blk, b2blk;
/*
 * First, choose kernel based on KB, which often can't be varied
 */
   for (i=0; i < ATL_AMM_NCASES && ATL_AMM_KBs[i] < K; i++);
/*
 * These kernels all take runtime M/N, and do well with near-square, so
 * blindly use this kernel with nM = CEIL(M/mu)*mu, nN = CEIL(N/nu)*nu
 */
   mu = ATL_AMM_MUs[i];
   nu = ATL_AMM_NUs[i];
   ku = ATL_AMM_KUs[i];
   nmu = (M+mu-1)/mu;
   nnu = (N+nu-1)/nu;
   nku = (K+ku-1)/ku;
   bM = nmu * mu;
   bN = nnu * nu;
   bK = nku * ku;
   #if ATL_AMM_MAXKMAJ > 1
      KK = ATL_AMM_KMAJOR(i) ? bK : K;
   #endif
/*
 * If alpha is one, then we only have to worry about beta
 */
   if (alpha == 1.0)
   {
      a2blk = (TA == AtlasNoTrans) ? ATL_AMM_AT2BLK_a1[i] : ATL_AMM_A2BLK_a1[i];
      b2blk = (TB == AtlasNoTrans) ? ATL_AMM_B2BLK_a1[i] : ATL_AMM_BT2BLK_a1[i];
      if (beta == ATL_rone)
         blk2c = ATL_AMM_BLK2C_a1_b1[i];
      else if (beta == ATL_rnone)
         blk2c = ATL_AMM_BLK2C_a1_bn[i];
      else if (beta == ATL_rzero)
         blk2c = ATL_AMM_BLK2C_a1_b0[i];
      else
         blk2c = ATL_AMM_BLK2C_a1_bX[i];
   }
/*
 * If C smallest matrix, apply alpha to it
 */
   else if (N > K && M > K)
   {
      alpC = alpha;
      a2blk = (TA == AtlasNoTrans) ? ATL_AMM_AT2BLK_a1[i] : ATL_AMM_A2BLK_a1[i];
      b2blk = (TB == AtlasNoTrans) ? ATL_AMM_B2BLK_a1[i] : ATL_AMM_BT2BLK_a1[i];
      if (alpha == ATL_rnone)
      {
         if (beta == ATL_rone)
            blk2c = ATL_AMM_BLK2C_an_b1[i];
         else if (beta == ATL_rnone)
            blk2c = ATL_AMM_BLK2C_an_bn[i];
         else if (beta == ATL_rzero)
            blk2c = ATL_AMM_BLK2C_an_b0[i];
         else
            blk2c = ATL_AMM_BLK2C_an_bX[i];
      }
      else
      {
         if (beta == ATL_rone)
            blk2c = ATL_AMM_BLK2C_aX_b1[i];
         else if (beta == ATL_rnone)
            blk2c = ATL_AMM_BLK2C_aX_bn[i];
         else if (beta == ATL_rzero)
            blk2c = ATL_AMM_BLK2C_aX_b0[i];
         else
            blk2c = ATL_AMM_BLK2C_aX_bX[i];
      }
   }
/*
 * If A smallest matrix, apply alpha to it
 */
   else if (K > N && M > N)
   {
      alpA = alpha;
      if (alpha == ATL_rnone)
         a2blk = (TA == AtlasNoTrans)?ATL_AMM_AT2BLK_an[i]:ATL_AMM_A2BLK_an[i];
      else
         a2blk = (TA == AtlasNoTrans)?ATL_AMM_AT2BLK_aX[i]:ATL_AMM_A2BLK_aX[i];
      b2blk = (TB == AtlasNoTrans) ? ATL_AMM_B2BLK_a1[i] : ATL_AMM_BT2BLK_a1[i];
      if (beta == ATL_rone)
         blk2c = ATL_AMM_BLK2C_a1_b1[i];
      else if (beta == ATL_rnone)
         blk2c = ATL_AMM_BLK2C_a1_bn[i];
      else if (beta == ATL_rzero)
         blk2c = ATL_AMM_BLK2C_a1_b0[i];
      else
         blk2c = ATL_AMM_BLK2C_a1_bX[i];
   }
/*
 * B is smallest matrix, apply alpha to it
 */
   else
   {
      alpB = alpha;
      a2blk = (TA == AtlasNoTrans) ? ATL_AMM_AT2BLK_a1[i] : ATL_AMM_A2BLK_a1[i];
      if (alpha == ATL_rnone)
         b2blk = (TB == AtlasNoTrans)?ATL_AMM_B2BLK_an[i]:ATL_AMM_BT2BLK_an[i];
      else
         b2blk = (TB == AtlasNoTrans)?ATL_AMM_B2BLK_aX[i]:ATL_AMM_BT2BLK_aX[i];
      if (beta == ATL_rone)
         blk2c = ATL_AMM_BLK2C_a1_b1[i];
      else if (beta == ATL_rnone)
         blk2c = ATL_AMM_BLK2C_a1_bn[i];
      else if (beta == ATL_rzero)
         blk2c = ATL_AMM_BLK2C_a1_b0[i];
      else
         blk2c = ATL_AMM_BLK2C_a1_bX[i];
   }
   #if ATL_AMM_MAXKMAJ > 1
      if (ATL_AMM_KMAJOR(i))
      {
         if (ATL_AMM_KRUNTIME(i))
            amm = ATL_AMM_KERN_b0[i];
         else
            amm = (ATL_AMM_KBs[i]==KK) ? ATL_AMM_KERN_b0[i]:ATL_AMM_KERN_K1[i];
      }
      else
   #endif
         amm = (bK == K) ? ATL_AMM_KERN_b0[i] : ATL_AMM_KERN_K1[i];
   #if ATL_AMM_MAXDUPB > 1
      szB = K*bN*ATL_AMM_DUPBN[i];
   #else
      szB = KK*bN;
   #endif
   vp = malloc((bM*bN + bM*KK + szB)*sizeof(TYPE) + 3*ATL_Cachelen);
   if (!vp)
      return(1);
   pB = ATL_AlignPtr(vp);
   pA = pB + szB;
   pA = ATL_AlignPtr(pA);
   pC = pA + bM*KK;
   pC = ATL_AlignPtr(pC);
/*
 * Copy A & B into workspace, and pad K its if necessary
 */
   a2blk(K, M, alpA, A, lda, pA);
   b2blk(K, N, alpB, B, ldb, pB);
   amm(nmu, nnu, KK, pA, pB, pC, pA, pB, pC);
   blk2c(M, N, alpC, pC, beta, C, ldc);

   free(vp);
   return(0);
}
