#include "atlas_misc.h"
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_rankK.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_rkkblk.h))
#include Mstr(Mjoin(Mjoin(atlas_,PRE),amm_rkkflag.h))


/*
 * This routine called when 2 < K <= MAXK
 */
int Mjoin(PATL,ammm_rkK)
(
   enum ATLAS_TRANS TA,
   enum ATLAS_TRANS TB,
   ATL_CSZT M,
   ATL_CSZT N,
   ATL_CSZT K,
   const SCALAR alpha,
   const TYPE *A,
   ATL_CSZT lda,
   const TYPE *B,
   ATL_CSZT ldb,
   const SCALAR beta,
   TYPE *C,
   ATL_CSZT ldc
)
{
   void *vp;
   TYPE *pC, *pA, *pB, *pA0, *c;
   const TYPE *a, *b;
   ammkern_t amm;
   ablk2cmat_t blk2c;
   cm2am_t a2blk, b2blk;
   size_t i, j, Mb, Nb, nmblks, nnblks, incAw, incC, incB, incAm, KK, szB;
   TYPE alpA=ATL_rone, alpB=ATL_rone;
   int mb, nb, mu, nu, ku, mbF, nbF, mbFU, nbFU, nmu, nnu, nmuF, nnuF;

   ATL_assert(K > 2 && K <= ATL_MAXK_RKK);
   amm = ATL_AMM_KERN_RKK[K-3];
   mu = ATL_AMM_MUs[K-3];
   nu = ATL_AMM_NUs[K-3];
   ku = ATL_AMM_KUs[K-3];

   mb = nb = 0;
   if (M < ATL_MAXM_RKK)
      mb = M+1;
   if (N < ATL_MAXN_RKK)
      nb = N+1;
   if (!mb && !nb)
   {
      i = mb = (ATL_MAXM_RKK/mu)*mu;
      j = nb = (ATL_MAXN_RKK/nu)*nu;
      while (2*i*j + K*(i+j) > ATL_L1elts)
      {
          if (i > j && i > 16)
             i -= mu;
          else if (j > 16)
             j -= nu;
          else break;
      }
      if (i > 15 && j > 15 && i >= (mu<<2) && j >= (nu<<2))
      {
         mb = i;
         nb = j;
      }
   }
   else if (!mb)  /* can only increase mb */
      mb = (ATL_MAXM_RKK/mu)*mu;
   else if (!nb)  /* can only increase nb */
      nb = (ATL_MAXN_RKK/nu)*nu;
//   printf("MB=%d, NB=%d, K=%d, mu=%d, nu=%d\n", mb, nb, (int)K, mu, nu);
   nmu = mb / mu;
   nnu = nb / nu;

   nmblks = M / mb;
   mbF = M - nmblks*mb;
/*
 * If mbF is small, add it to last full block and make oversized block
 */
   if (mbF < mu+mu && nmblks)
   {
      nmblks--;
      mbF += mb;
   }
   Mb = nmblks*mb;
   nmuF = (mbF+mu-1)/mu;
   mbFU = nmuF * mu;
   nnblks = N / nb;
   nbF = N - nnblks*nb;
   if (nbF < nu+nu && nnblks)
   {
      nnblks--;
      nbF += nb;
   }
   Nb = nnblks*nb;
   nnuF = ((nbF+nu-1)/nu);
   nbFU = nnuF*nu;

   if (beta == ATL_rone)
      blk2c = ATL_RKK_BLK2C_a1_b1[K-3];
   else if (beta == ATL_rnone)
      blk2c = ATL_RKK_BLK2C_a1_bn[K-3];
   else if (beta == ATL_rzero)
      blk2c = ATL_RKK_BLK2C_a1_b0[K-3];
   else
      blk2c = ATL_RKK_BLK2C_a1_bX[K-3];
/*
 * If alpha is one, just copy without scale
 */
   if (alpha == 1.0)
   {
      a2blk = (TA == AtlasNoTrans) ?
               ATL_RKK_AT2BLK_a1[K-3] : ATL_RKK_A2BLK_a1[K-3];
      b2blk = (TB == AtlasNoTrans) ?
              ATL_RKK_B2BLK_a1[K-3] : ATL_RKK_BT2BLK_a1[K-3];
   }
/*
 * Otherwise, if A smallest matrix, apply alpha to it
 */
   else if (N > M)
   {
      alpA = alpha;
      if (alpha == ATL_rnone)
         a2blk = (TA == AtlasNoTrans) ?
                 ATL_RKK_AT2BLK_an[K-3] : ATL_RKK_A2BLK_an[K-3];
      else
         a2blk = (TA == AtlasNoTrans) ?
                 ATL_RKK_AT2BLK_aX[K-3] : ATL_RKK_A2BLK_aX[K-3];
      b2blk = (TB == AtlasNoTrans) ?
              ATL_RKK_B2BLK_a1[K-3] : ATL_RKK_BT2BLK_a1[K-3];
   }
/*
 * Otherwise, B is smallest matrix, apply alpha to it
 */
   else
   {
      alpB = alpha;
      a2blk = (TA == AtlasNoTrans) ?
              ATL_RKK_AT2BLK_a1[K-3] : ATL_RKK_A2BLK_a1[K-3];
      if (alpha == ATL_rnone)
         b2blk = (TB == AtlasNoTrans) ?
                 ATL_RKK_B2BLK_an[K-3] : ATL_RKK_BT2BLK_an[K-3];
      else
         b2blk = (TB == AtlasNoTrans) ?
                 ATL_RKK_B2BLK_aX[K-3] : ATL_RKK_BT2BLK_aX[K-3];
   }
/*
 * Allocate workspace
 */
   i = Mmax(mb, mbFU);
   j = Mmax(nb, nbFU);
   KK = K;
   #if ATL_AMM_MAXKMAJ > 1
      if (ATL_AMM_KMAJOR(K-3))
         KK = ((K+ku-1)/ku)*ku;
   #endif
   szB = j*KK;
   vp = malloc((i*j + szB + (Mb+mbFU)*KK)*sizeof(TYPE) + 3*ATL_Cachelen);
   if (!vp)
      return(1);
   pA0 = pA = ATL_AlignPtr(vp);
   pB = pA + K*(Mb+mbFU);
   pB = ATL_AlignPtr(pB);
   pC = pB + szB;
   pC = ATL_AlignPtr(pC);
/*
 * Do last nbF-wide column of C & B
 * NOTE: Rewrite to do have mb0 be last block, and do column panels
 * right-to-left
 */
   if (TB == AtlasNoTrans)
   {
      b = B + Nb*ldb;
      incB = nb*ldb;
   }
   else
   {
       b = B + Nb;
       incB = nb;
   }
   C += ldc*Nb;
   incC = ldc*nb;
   a = A;
   incAm = (TA == AtlasNoTrans) ? mb : mb*lda;
   incAw = mb*KK;
   pA = pA0;
   b2blk(K, nbF, alpB, b, ldb, pB);
   b -= incB;
   c = C;
/*
 * Loop over all standard-sized MB blocks of A, copying as we do gemm
 */
   for (i=nmblks; i; i--)
   {
      TYPE *pAn = pA + incAw;
      a2blk(K, mb, alpA, A, lda, pA);
      amm(nmu, nnuF, K, pA, pB, pC, pAn, pB, pC);
      blk2c(mb, nbF, ATL_rone, pC, beta, c, ldc);
      pA = pAn;
      A += incAm;
      c += mb;
   }
   a2blk(K, mbF, alpA, A, lda, pA);
   amm(nmuF, nnuF, K, pA, pB, pC, pA0, pB, pC);
   blk2c(mbF, nbF, ATL_rone, pC, beta, c, ldc);
   C -= incC;
   for (j=nnblks; j; j--)
   {
      pA = pA0;
      c = C;
      b2blk(K, nb, alpB, b, ldb, pB);
/*
 *    Loop over all standard-sized MB blocks of A, copying as we do gemm
 */
      for (i=nmblks; i; i--)
      {
         TYPE *pAn = pA + incAw;
         amm(nmu, nnu, K, pA, pB, pC, pAn, pB, pC);
         blk2c(mb, nb, ATL_rone, pC, beta, c, ldc);
         pA = pAn;
         c += mb;
      }
      amm(nmuF, nnu, K, pA, pB, pC, pA0, pB, pC);
      blk2c(mbF, nb, ATL_rone, pC, beta, c, ldc);
      b -= incB;
      C -= incC;
   }

   free(vp);
   return(0);
}
